//--------------------------------------------------- 2014-01-20 ----------
// Arduino回転計シールド    
//------------------------------------ written by iizuka --- Japan --------
#include <MsTimer2.h>

const byte Digit_N = 0xFF;
const byte Digit_0 = 0x03;
const byte Digit_1 = 0x9F;
const byte Digit_2 = 0x25;
const byte Digit_3 = 0x0D;
const byte Digit_4 = 0x99;
const byte Digit_5 = 0x49;
const byte Digit_6 = 0x41;
const byte Digit_7 = 0x1F;
const byte Digit_8 = 0x01;
const byte Digit_9 = 0x09;

const byte STATUS_READY   = 0;
const byte STATUS_START   = 1;
const byte STATUS_DISPLAY = 2;

#define LED_SEG_A       19   // Port Analog 5
#define LED_SEG_B       18   // Port Analog 4
#define LED_SEG_C       17   // Port Analog 3
#define LED_SEG_D       16   // Port Analog 2
#define LED_SEG_E       15   // Port Analog 1
#define LED_SEG_F       14   // Port Analog 0
#define LED_SEG_G        4   // Port Analog 4
#define LED_SEG_DP      11   // Port Digial 11

#define LED_DIG_5        6   // Port Digial 6
#define LED_DIG_4        7   // Port Digial 7 
#define LED_DIG_3        8   // Port Digial 8
#define LED_DIG_2        9   // Port Digial 9
#define LED_DIG_1       10   // Port Digial 10

#define SENSOR_1         2   // Port Digial 2
#define SENSOR_2         5   // Port Digial 5

#define RELAY            3   // Port Digital 3  リレーは LOW でＯＮ 
#define SW1              0   // Port Digital 0
#define ENCODER_A        1   // Port Digital 1
#define ENCODER_B        5   // Port Digital 5
#define LED1            13   // Port Digital 13
#define LED2            12   // Port Digital 12

byte led_val[5];
byte led_pos;

volatile unsigned int count;
volatile byte stat;
volatile unsigned long start_micros;
volatile unsigned long end_micros;
volatile unsigned long rotate;

unsigned long out_rpm;
unsigned long end_millis;

//--------------------------------------------------------------------------
// セグメント点灯パターンをポートに出力
// 　この手の処理は８ビットをまとめてポートに出力するのが定石だが
//   arduinoの素直な使い方では１ビット単位出力(digitalWrite)を使う
//   高速処理したいならPORTB,PORTD等にバイト単位で書き込めばよいが
//   POERBを使えばSPIは使えないしPORTDを使えばSerial通信は使えなくなる
//   といった制約を受ける
// 参考
//   PORTB - Digital8-13   PORTD - Digital0-7  PORTC - Analog0-5
//--------------------------------------------------------------------------
void led_seg_out(byte seg_data ) {
  digitalWrite(LED_SEG_A, bitRead( seg_data, 7 ));  
  digitalWrite(LED_SEG_B, bitRead( seg_data, 6 ));  
  digitalWrite(LED_SEG_C, bitRead( seg_data, 5 ));  
  digitalWrite(LED_SEG_D, bitRead( seg_data, 4 ));  
  digitalWrite(LED_SEG_E, bitRead( seg_data, 3 ));  
  digitalWrite(LED_SEG_F, bitRead( seg_data, 2 ));  
  digitalWrite(LED_SEG_G, bitRead( seg_data, 1 ));  
}

//--------------------------------------------------------------------------
// ７セグメントＬＥＤが同時に点灯しているように見せかける処理
//    2msecごとにこの処理が行われる
//--------------------------------------------------------------------------
void strobe() {
  switch( led_pos ) {
    case 1:
      digitalWrite(LED_DIG_5,HIGH);  // LED OFF
      led_seg_out(led_val[4]);
      digitalWrite(LED_DIG_1,LOW);   // LED ON 
      led_pos++;
      break;
    case 2:
     digitalWrite(LED_DIG_1,HIGH);  
     led_seg_out(led_val[3]);
     digitalWrite(LED_DIG_2,LOW);    
     led_pos++;
     break;
    case 3:
     digitalWrite(LED_DIG_2,HIGH);    
     led_seg_out(led_val[2]);
     digitalWrite(LED_DIG_3,LOW);    
     led_pos++;
     break;
    case 4:
     digitalWrite(LED_DIG_3,HIGH);    
     led_seg_out(led_val[1]);
     digitalWrite(LED_DIG_4,LOW);    
     led_pos++;
     break;
    case 5:
     digitalWrite(LED_DIG_4,HIGH);    
     led_seg_out(led_val[0]);
     digitalWrite(LED_DIG_5,LOW);    
     led_pos = 1;
     break;
  }
  
  count++;
}

//--------------------------------------------------------------------------
// １桁数値を７セグＬＥＤの点灯パターンに変換する
//--------------------------------------------------------------------------
byte ByteToDigit( byte number ) {
  byte return_value;
  
  switch( number ) {
    case 0: return_value = Digit_0; break;
    case 1: return_value = Digit_1; break;
    case 2: return_value = Digit_2; break;
    case 3: return_value = Digit_3; break;
    case 4: return_value = Digit_4; break;
    case 5: return_value = Digit_5; break;
    case 6: return_value = Digit_6; break;
    case 7: return_value = Digit_7; break;
    case 8: return_value = Digit_8; break;
    case 9: return_value = Digit_9; break;
  }
  return return_value;
} 

//--------------------------------------------------------------------------
// 回転数から７セグＬＥＤの点灯パターンを生成する
// sprintfで文字列に変換してから７セグＬＥＤの点灯パターンに変換する方が
// 単純になるがarduinoの標準機能のみで処理するようにしている
//--------------------------------------------------------------------------
void LongToDigit( unsigned long rpm ) {
  byte val[5];
 
  val[0] = Digit_0;
  val[1] = Digit_N;
  val[2] = Digit_N;
  val[3] = Digit_N;
  val[4] = Digit_N;
  for( int i = 0 ; i < 5 ; i++ ) {
    if( rpm != 0 ) {
      val[i] = ByteToDigit( (byte)(rpm % 10) );
    }
    rpm = rpm / 10;
  }
  for( int i = 0 ; i < 5 ; i++ ) {
    led_val[i] = val[i];  // 最終結果は配列 led_val[] に代入される
                          // 一気に書き換えないと前回の回転数と今回の回転数が合成表示される時間が長くなる
                          // 上の処理は割り算や剰余計算があるので処理に時間がかかる
  }
}

//--------------------------------------------------------------------------
// 回転センサー割込処理　ここで回転数を計測
//  Digital2ピンの外部割込みを利用
//--------------------------------------------------------------------------
void interrupt_rotate()
{
  unsigned long rpm;

  switch( stat ) {
    case STATUS_READY:
      stat = STATUS_START;
      start_micros = micros();   // 計測開始
      rotate = 0;
      count = 0;
      break;
    case STATUS_START:
      rotate++;
      if( count >= 500 ) {     // 約１秒経過したら
        end_micros = micros();  // 計測終了
        if( start_micros < end_micros ) {
          rpm = rotate * 60000L / ((end_micros - start_micros) / 1000);
            //  回転数 × 60000[msec/60秒] / ((計測終了時刻 - 計測開始時刻) / 1000 [microsec/msec])
        }
        LongToDigit( rpm );
        out_rpm = rpm;
        stat = STATUS_READY;
      }
      break;
  }      
}

//--------------------------------------------------------------------------
// arduinoセットアップ
//--------------------------------------------------------------------------
void setup()   {
  pinMode( LED_SEG_A, OUTPUT );
  pinMode( LED_SEG_B, OUTPUT );
  pinMode( LED_SEG_C, OUTPUT );
  pinMode( LED_SEG_D, OUTPUT );
  pinMode( LED_SEG_E, OUTPUT );
  pinMode( LED_SEG_F, OUTPUT );
  pinMode( LED_SEG_G, OUTPUT );

  pinMode( LED_DIG_5, OUTPUT );
  pinMode( LED_DIG_4, OUTPUT );
  pinMode( LED_DIG_3, OUTPUT );
  pinMode( LED_DIG_2, OUTPUT );
  pinMode( LED_DIG_1, OUTPUT );

  pinMode( RELAY,     OUTPUT );
  digitalWrite( RELAY, HIGH );
  pinMode( LED1,      OUTPUT );
  pinMode( LED2,      OUTPUT );
  
  pinMode( SW1,       INPUT  );
  pinMode( ENCODER_A, INPUT  );
  pinMode( ENCODER_B, INPUT  );

  led_val[4] = Digit_8;
  led_val[3] = Digit_8;
  led_val[2] = Digit_8;
  led_val[1] = Digit_8;
  led_val[0] = Digit_8;

  Serial.begin(9600);
  end_millis = millis() + 1000;

  count = 0;
  stat = STATUS_READY;
  attachInterrupt(0, interrupt_rotate, FALLING);  // on digital pin 2

  led_pos = 1;  
  MsTimer2::set(2, strobe);
  MsTimer2::start(); 
}

//--------------------------------------------------------------------------
// arduinoループ
//--------------------------------------------------------------------------
void loop()
{
  // 回転していないときに表示をゼロにする処理 殆どの処理は割り込み処理で行われる
  if( count > 1000 ) {
    led_val[0] = Digit_0;  
    led_val[1] = Digit_N;
    led_val[2] = Digit_N;
    led_val[3] = Digit_N;
    led_val[4] = Digit_N;
    count = 0;
    out_rpm = 0;
    stat = STATUS_READY;
  }
  
  // シリアルポートに回転数を出力
  if( end_millis < millis() ) {
    Serial.println( out_rpm );
    end_millis += 1000;
  }
/*  
  if( digitalRead( SW1 ) == HIGH ) {
    digitalWrite( LED1, HIGH );
    digitalWrite( LED2, LOW  );
  } else {
    digitalWrite( LED1, LOW  );
    digitalWrite( LED2, HIGH );  
  }
*/  
}

//--------- E N D -------------------------------------------------------
